;============================================================
; Zelda Dungeon Expansion Patch
; -------------------------------
; v2.0
; by gzip
;
; There are several sections that are meant to be edited:
;
; 1) Data for Dungeon Blocks (optional)
; 2) Pointers for Dungeon Column Groups (optional)
; 3) New Column Definitions
; 4) New Dungeon Screens
;============================================================

;============================================================
; Dungeon Blocks
; --------------
; Remaps the dungeon blocks and modify the block index mask
; to support 16 possible blocks instead of 8
;============================================================

; Pointer and data for dungeon blocks
; --------------

.PATCH $16792 ; DungeonBlocksPointer
  .word $93F0 ; was A718 $18, $A7

; Data for Dungeon Blocks (optionally edit the new blocks data section)
; --------------

; Make sure this matches the pointer above
.PATCH 05:93F0 ; 15400 DungeonBlocksData Add 8 blocks for a total of 16
  .db $B0, $74, $94, $B4, $70 ,$68, $F4, $24 ; existing blocks data ($00-$07)
  .db $40, $44, $48, $4C, $50, $54, $58, $5C ; new blocks data ($08-$0F)

; Mask for dungeon block index
; --------------

.PATCH 05:A77E ; 1678E
  AND #$0F ; get the square index with a max of 15, was #$07


;============================================================
; Dungeon Columns
; ---------------
; This remaps the pointer table for the dungeon column groups
; and adds two additional groups with option for more
;============================================================

; Pointer for dungeon columns pointer table
; --------------

.PATCH 05:A755 ; 16765 DungeonColumnsPointerTableLo
  .word $9352 ; was $A704 (04 A7)

.PATCH 05:A75A ; 1676A DungeonColumnsPointerTableHi
  .word $9353 ; was $A705 (05 A7)

; Pointers for Dungeon Column Groups (optionally editable)
; --------------

; Make sure this matches the pointer above at A755
.PATCH 05:9352 ; 15362 DungeonColumnsPointerTable
  .word $A2D6 ; existing group 0 at 162E6
  .word $A2E6 ; existing group 1 at 162F6
  .word $A304 ; existing group 2 at 16314
  .word $A315 ; existing group 3 at 16325
  .word $A326 ; existing group 4 at 16336
  .word $A33D ; existing group 5 at 1634D
  .word $A354 ; existing group 6 at 16364
  .word $A365 ; existing group 7 at 16375
  .word $A377 ; existing group 8 at 16387
  .word $A391 ; existing group 9 at 163A1
  .word $ADE0 ;      new group A at 16DF0
  .word $AE10 ;      new group B at 16E20
  ; reserve address space for more groups
  .word $0000 ;      new group C at ?????
  .word $0000 ;      new group D at ?????
  .word $0000 ;      new group E at ?????
  .word $0000 ;      new group F at ?????

; New Column Definitions
; --------------

; Make sure this matches the group A pointer above
.PATCH 05:ADE0 ; 16DF0 DungeonColumnGroupA
  .db $81, $16, $01, $16, $01 ; A0
  .db $96, $21, $16 ; A1
  .db $91, $10, $01, $00, $01 ; A2

; Make sure this matches the group B pointer above
.PATCH 05:AE10 ; 16E20 DungeonColumnGroupB
  .db $88, $09, $0A, $0B, $0C, $0D, $0E ; B0 A test column containing the new blocks


;============================================================
; Code Changes
; ---------------
; Determine how blocks are written and if they are walkable
;============================================================

; Pointers for the two types of blocks
; --------------

WriteBlockSequence := $A7C9 ; C0 C1 C2 C3
WriteTileRepeat    := $A7E0 ; C0 C0 C0 C0

.PATCH 05:A786 ; 16796 in LoopSquareRow
  JSR DetermineBlockType

; Draw one block type or the other based on the tile id found in DungeonBlocksData
; --------------

.PATCH 05:9372 ; 15382
  DetermineBlockType:
    CMP #$30    ; if index is under $30
    BCC ++      ; then repeat the tile 4x
    CMP #$60    ; if index is under $60 (between $30 and $5F)
    BCC +       ; then write a block sequence
    CMP #$70    ; if index is under $70 (between $60 and $6F)
    BCC ++      ; then repeat the tile 4x
    CMP #$F1    ; if index is over $F1
    BCS ++      ; then repeat the tile 4x
  + JMP WriteBlockSequence
 ++ JMP WriteTileRepeat

; Dungeon Screens Pointer
; -----------------------
; Expand available screens by changing the screens pointer based on screen index

.PATCH 05:A723 ; 16733 Map generator checks this code
  JMP SetDungeonScreenPointer
  NOP
  StoreDungeonScreenPointer:
  STX $02
  STY $03
  PHA           ; put adjusted index on the stack as expected

.PATCH 05:BBB5 ; 17BC5 Map generator reads the pointers from this code
  SetDungeonScreenPointer:
    CMP #$2A    ; compare to default max
    BCS +       ; set alternate pointer if over
    LDX #$DE    ; otherwise set the default pointer
    LDY #$A0
    JMP ++      ; we're done so return
  + LDX #$62    ; set the alternate pointer
    LDY #$AC
    SBC #$2A    ; readjust the index to be zero based
    ; store the pointer at the expected address when done
 ++ JMP StoreDungeonScreenPointer

; New Dungeon Screens (editable)
; --------------
; Add new dungeon screens. Use as-is or change as needed

.PATCH 05:AC62 ; 16C72, make sure this matches the alternate pointer above
  .data $A0, $A1, $70, $70, $A1, $A0, $A0, $A1, $70, $70, $A1, $A0 ; $2A
  .data $00, $1B, $00, $11, $00, $1B, $1B, $00, $11, $00, $1B, $00 ; $2B
  .data $00, $00, $00, $22, $02, $A2, $15, $12, $22, $00, $00, $00 ; $2C
  .data $00, $00, $31, $31, $00, $12, $12, $00, $31, $31, $00, $00 ; $2D
  .data $00, $78, $0A, $09, $00, $0B, $0B, $00, $09, $0A, $78, $00 ; $2E
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $2F
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $30
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $31
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $32
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $33
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $34
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $35
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $36
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $37
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $38
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $39
  .data $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00, $00 ; $3A

; 1623A Level 1 room to right of entrance
