;============================================================
; Zelda Overworld Expansion Patch
; -------------------------------
; v2.0
; by gzip
;
; Much of the code in this file shouldn't be changed but
; there are several sections that are meant to be edited.
; They are marked "editable":
;
; 1) New Tile Sequences
; 2) New Sample Columns
; 3) Additional Screens
; 4) Tile/Block Walkability Changes (only a few bytes are editable)
;============================================================

;============================================================
; Tile Data
; ---------
; This data determines which tiles get rendered for each block.
;============================================================

; Tile Codes
; ----------
; Tile Sequences for Block Ids >= 0x10
; This pointer gets shifted up $10 bytes in order to free up another $10 (16) tile codes to be used at the end

.PATCH 05:AA62 ; 16A72
  LDA TileCodesOW - $10,X ; Tile code pointer

.PATCH 05:A97C ; $1698C
  TileCodesOW:
    .data $03, $04, $70, $C8, $BC, $8D, $8F, $93, $95, $C4, $CE, $D8, $B0, $B4, $AA, $AC ; blocks $10-$1F
    .data $B8, $9C, $A6, $9A, $A2, $A0, $E5, $E6, $E7, $E8, $E9, $EA, $C0, $E0, $78, $7A ; blocks $20-$2F
    .data $7E, $80, $CC, $D0, $D4, $DC, $89, $84 ; blocks $30-$37
    ; the following blocks use the space freed up by shifting up the pointer
    .data $30, $34, $38, $3C, $40, $44, $48, $4C ; blocks $38-$3F unwalkable blocks
    .data $50, $54, $58, $5C ; blocks $40-$43 walkable blocks
    .data $24, $F4, $F8, $FC ; blocks $44-$47 reserved for future use

; move the displaced data
.PATCH 05:A3E6 ; 163F6, this is where the compressed dungeon passage columns were located so reuse the space
  TileCodesUnder0x10:
    .data $24, $6F, $F3, $FA, $98, $90, $8F, $95, $8E, $90, $74, $76, $F3, $24, $26, $89

.PATCH 05:ABF7 ; 16C07 repoint to new location
  CMP TileCodesUnder0x10,X ; was A97C,X

; Vanilla Tile Sequences
; ----------------------
; Tile Sequences for Block Ids < 0x10
; Unchanged, included for convenience

.PATCH 05:A9B4 ; $169C4
  .data $24, $24, $24, $24 ; 00 - Dungeons Passages Top of Stairs (black)
  .data $6F, $6F, $6F, $6F ; 01 - Dungeons Passages Stairs
  .data $F3, $F3, $F3, $F3 ; 02 - Dungeons Passages Background (blocked)
  .data $FA, $FA, $FA, $FA ; 03 - Dungeons Passages Walls (bricks)
  .data $98, $95, $26, $26 ; 04 - Unused <-- (editable)
  .data $90, $95, $90, $95 ; 05 - Ocean
  .data $8F, $90, $8F, $90 ; 06 - Shore top edge
  .data $95, $96, $95, $96 ; 07 - Shore bottom edge
  .data $8E, $93, $90, $95 ; 08 - Shore left edge
  .data $90, $95, $92, $97 ; 09 - Shore right edge
  .data $74, $74, $75, $75 ; 0A - Mountain Stairs
  .data $76, $77, $76, $77 ; 0B - Bridge
  .data $F3, $24, $F3, $24 ; 0C - Dungeons Passages Walkable Space (black) / Cave Entrance in OW
  .data $24, $24, $24, $24 ; 0D - Overworld Caves Ground
  .data $26, $26, $26, $26 ; 0E - Overworld Ground
  .data $89, $88, $8B, $88 ; 0F - Waterfall

; New Tile Sequences (editable)
; ------------------
; These are example sequences that can be modified as needed

.PATCH 05:AD50 ; 16D60
  BlockSequencesAbove0x47:
    ; these blocks replace defaults in order to reclaim some tiles
    ; *** Don't edit these
    .data $78, $79, $26, $26 ; block $48, replaces $2E
    .data $26, $26, $7C, $7D ; block $49, replaces $2F
    .data $7E, $7F, $26, $26 ; block $4A, replaces $30
    .data $26, $26, $82, $83 ; block $4B, replaces $31
    .data $9A, $9B, $26, $9D ; block $4C, replaces $23
    .data $26, $9D, $9E, $9F ; block $4D, replaces $21
    .data $AA, $AB, $26, $26 ; block $4E, replaces $1E
    .data $26, $26, $AE, $AF ; block $4F, replaces $1F
    ; *** These tiles can be changed as needed
    .data $99, $9C, $99, $9C ; block $50
    .data $AC, $AD, $AC, $AD ; block $51
    .data $E4, $E4, $E4, $E4 ; block $52
    .data $80, $81, $80, $81 ; block $53
    .data $9E, $9F, $9A, $9B ; block $54
    .data $A4, $A5, $A0, $A1 ; block $55
    .data $9D, $A2, $9D, $A2 ; block $56
    .data $9D, $A2, $26, $9D ; block $57
    .data $26, $9D, $9D, $A2 ; block $58
    .data $9D, $9D, $9D, $9D ; block $59
    .data $A3, $A3, $A3, $A3 ; block $5A
    .data $7A, $7A, $7B, $7B ; block $5B
    .data $FE, $FD, $FE, $FD ; block $5C
    .data $8D, $94, $91, $98 ; block $5D
    .data $8D, $94, $8F, $96 ; block $5E
    .data $8F, $96, $91, $98 ; block $5F
    .data $8F, $96, $8F, $96 ; block $60
    .data $93, $8E, $97, $92 ; block $61
    .data $68, $68, $68, $68 ; block $62
    .data $F8, $F8, $F8, $F8 ; block $63
    .data $FA, $FA, $FA, $FA ; block $64
    .data $26, $26, $FA, $FA ; block $65
    .data $FA, $FA, $26, $26 ; block $66
    .data $FA, $26, $FA, $26 ; block $67
    .data $26, $FA, $26, $FA ; block $68
    .data $6F, $6F, $6F, $6F ; block $69
    .data $10, $23, $02, $00 ; block $6A
    .data $12, $19, $02, $01 ; block $6B


;============================================================
; Column Data
; -----------
; This data determines which blocks are used in which columns.
;============================================================

; Column Groups Pointer Table
; ----------------------------
; These must match the column data which follows

.PATCH 06:9D0F ; 19D1F OverworldColumnsPointerTable
  ; existing groups that have been regrouped to be consolidated
  .word $9BD8 ; group 0 at 15BE8
  .word $9C31 ; group 1 at 15C41
  .word $9C95 ; group 2 at 15CA5
  .word $9CF0 ; group 3 at 15D00
  .word $9D56 ; group 4 at 15D66
  .word $9DC2 ; group 5 at 15DD2
  .word $9E35 ; group 6 at 15E45
  .word $9EA3 ; group 7 at 15EB3
  .word $9F09 ; group 8 at 15F19
  .word $9F74 ; group 9 at 15F84
  ; new groups start here, make sure these pointers match the locations below
  .word $B8A0 ; new group A at 178B0
  .word $B930 ; new group B at 17940
  .word $B9B0 ; new group C at 179C0
  .word $BA50 ; new group D at 17A60
  .word $BAA0 ; new group E at 17AB0
  .word $BB10 ; new group F at 17B20

; Default Columns
; ---------------
; These are regrouped to maximize the number of columns in each group,
; thereby freeing up more space for new columns

.PATCH 05:9BD8 ; groups 0-9 at 15BE8
  .incbin "column_data_regrouped.bin"

; New Sample Columns (editable)
; ------------------
; Use them as-is is or replace/remove/augment them
; Make sure these pointers match the locations in the
; column groups pointer tables above

.PATCH 05:B8A0 ; new group A at 178B0 (bushes)
.data $99, $19, $0E, $0E, $19, $19, $19, $0E, $0E
.data $99, $19, $0E, $0E, $19, $28, $19, $0E, $0E
.data $99, $19, $0E, $0E, $19
.data $99, $19, $19, $19, $19, $19, $19, $0E, $0E
.data $99, $19, $19, $19, $19, $0E, $19
.data $99, $19, $19, $19, $28, $0E, $19
.data $99, $19, $19, $19, $19, $19, $19, $19, $0E
.data $99, $19, $0E, $0E, $0E, $28, $0E, $0E, $0E, $19, $19

.PATCH 05:B930 ; new group B at 17940 (rocks)
.data $9B, $1B, $0E, $0E, $13, $13, $13, $0E, $0E, $1A, $1B
.data $9B, $1B, $0E, $0E, $13, $26, $13, $0E, $0E, $1A, $1B
.data $9B, $1B, $0E, $13, $13, $13, $13, $13, $0E, $1A, $1B

.PATCH 05:B9B0 ; new group C at 179C0 (rivers)
.data $9B, $1B, $0E, $0E, $15, $17, $0E, $0E, $0E, $1A, $1B
.data $9B, $1B, $0E, $0E, $06, $05, $08, $08, $08, $08, $08
.data $9B, $1B, $0E, $0E, $15, $08, $08, $08, $08, $08, $08
.data $9B, $1B, $0E, $0E, $06, $05, $09, $09, $09, $09, $09
.data $9B, $1B, $0E, $0E, $06, $07, $0E, $0E, $0E, $1A, $1B
.data $9B, $1B, $0E, $0E, $06, $07, $0E, $0E, $0E, $0A, $0A
.data $9B, $1B, $0E, $0E, $06, $05, $08, $08, $08, $08, $08
.data $9B, $1B, $0E, $0E, $16, $09, $09, $09, $09, $09, $09
.data $9B, $1B, $0E, $0E, $06, $05, $09, $09, $09, $09, $09
.data $9B, $1B, $0E, $0E, $16, $18, $0E, $0E, $0E, $1A, $1B

.PATCH 05:BA50 ; new group D at 17A60 (bridge)
.data $88, $08, $08, $08, $08, $0B, $08, $08, $08, $08, $08
; The sample columns in add_test_column.ips get added here

.PATCH 05:BAA0 ; new group E at 17AB0 (trees)
.data $9B, $1B, $0E, $1C, $4F, $0E, $1C, $4F, $0E, $1A, $1B
.data $9B, $1B, $0E, $4E, $20, $0E, $4E, $20, $0E, $1A, $1B
.data $9B, $1B, $0E, $4E, $20, $1C, $4F, $0E, $0E, $1A, $1B
.data $9B, $1B, $0E, $0E, $0E, $1C, $4F, $0E, $0E, $1A, $1B
.data $9B, $1B, $0E, $0E, $0E, $4E, $20, $0E, $0E, $1A, $1B
.data $9B, $1B, $0E, $1C, $4F, $4E, $20, $0E, $0E, $1A, $1B

.PATCH 05:BB10 ; new group F at 17B20 (armos and graves)
.data $9B, $1B, $0E, $2C, $0E, $2C, $0E, $2C, $0E, $1A, $1B
.data $9B, $1B, $0E, $0E, $14, $0E, $14, $0E, $0E, $1A, $1B
.data $9B, $1B, $0E, $0E, $29, $0E, $14, $0E, $0E, $1A, $1B
.data $9B, $1B, $0E, $0E, $0E, $14, $0E, $0E, $0E, $1A, $1B
.data $9B, $1B, $0E, $0E, $0E, $29, $0E, $0E, $0E, $1A, $1B


;============================================================
; Screens Data
; ------------
; Update the screens data to reflect new columns groupings
;============================================================

; Overworld Screens Data
; ----------------------
; New screens data with updated column references to reflect the new grouping

.PATCH 05:9418 ; remapped screens data at 15428
  .incbin "screens_data.bin"

; Additional Screens (editable)
; ------------------
; Four additional overworld screens are available

.PATCH 05:859E ; $145AE
  ScreensPointerAlt:
    .data $AFC0 ; Make sure this matches the pointer below

.PATCH 05:AFC0 ; 16FD0 Make sure this matches the pointer above
  .data $01, $65, $02, $02, $02, $02, $60, $11, $11, $61, $02, $02, $02, $02, $64, $01 ; screen $7C
  .data $01, $65, $02, $02, $02, $02, $60, $11, $11, $61, $02, $02, $02, $02, $64, $01 ; screen $7D
  .data $01, $65, $02, $02, $02, $02, $60, $11, $11, $61, $02, $02, $02, $02, $64, $01 ; screen $7E
  .data $01, $65, $02, $02, $02, $02, $60, $11, $11, $61, $02, $02, $02, $02, $64, $01 ; screen $7F

; Passage Screens Data
; --------------------
; Uncompressed and repoint dungeon passage columns which share overworld columns rendering code

.PATCH 05:AB3A ; 16B4A
  LDA #$C9 ; DungeonPassageColumnsLo
  LDX #$BB ; DungeonPassageColumnsHi

; Make sure this matches the pointer above
.PATCH 05:BBC9 ; 17BD9
  DungeonPassageColumns:
    .data $82, $03, $03, $03, $03, $03, $03, $02, $0C, $03, $03
    .data $80, $01, $01, $01, $01, $01, $01, $01, $01, $03, $03
    .data $82, $03, $03, $02, $02, $0C, $01, $01, $01, $03, $03
    .data $82, $03, $03, $02, $02, $0C, $03, $02, $0C, $03, $03
    .data $82, $03, $03, $03, $03, $03, $03, $03, $03, $03, $03


;============================================================
; Code Changes
; ------------
; Various logic changes to make it all work
;============================================================

; Vanilla Variables
; -----------------
; These should not change

BlockCode            := $0A
BlockIndex           := $0D
LevelNum             := $10
SaveSlot             := $16
CurrentScreen        := $EB
CurrentTile          := $049E
QuestNum             := $062D
ScreenAttrs_3        := $69FE
ScreenAttrs_5        := $6AFE
ScreensPointer       := $9F9C
TileCodesSecretsOW   := $A976
NextSquare           := $AAA4
CheckSpecialBlocks   := $AABF
MapBlockIdAbove0x0F  := $AAF7
WriteLastTile        := $AB0A
MapBlockIdUnder0x10  := $AB0E
SetTile              := $EE90


; Hack Variables
; --------------
; These can change if there's a conflict

ColumnGroup          := $628 ; unused according to datacrystal

; Block Replacement Data
; ----------------------
; Reclaim several tiles by remapping blocks with wasted ground tiles

.PATCH 05:A3D4 ; 163E4, this is where the compressed dungeon passage columns were located so reuse the space
  ReplacementBlocks:
    .data $2E, $2F, $30, $31, $23, $21, $1E, $1F ; replace these blocks
    .data $48, $49, $4A, $4B, $4C, $4D, $4E, $4F ; with these tile sequences
    .data $FF, $FF ; mark free space

.PATCH 05:AA89 ; 16A99
  JSR CheckReplacements ; was CheckSecretTileReplacement

.PATCH 05:BB60 ; 17B70 map generator checks this location
  CheckReplacements:
    ; exit if block is not between 1E and 31
    LDX BlockIndex
    CPX #$1E
    BCC ++
    CPX #$32
    BCS ++
    ; loop through the 8 blocks above to check for a match
    LDX #$07
  - LDY ReplacementBlocks,X  ; map generator checks this location as well
    CPY BlockIndex
    BEQ +                    ; jump if the block is found
    DEX
    BPL -
    BMI ++                   ; jump to secret check if nothing was found
    ; replace the block
  + PHA
    TXA
    ADC #$07 ; index to the replacement
    TAX
    PLA
    LDY ReplacementBlocks,X  ; load the replacement value
    STY BlockIndex           ; and replace it
    RTS
 ++ JSR CheckSpecialBlocks
    RTS

; Compression Hijacks
; -------------------
; This code enables compression for vanilla column groups,
; but leaves it disabled for extended column groups in order to gain more blocks

.PATCH 05:AA38 ; 16A48
  JMP StoreColumnGroup
  ExitStore:

.PATCH 05:AA5D ; 16A6D
  JMP MaskBlock
  NOP          ; Need to take up 4 bytes
  ExitMask:

.PATCH 05:AA9A ; 16AAA
  JMP CheckBlockRepeat
  NOP          ; Need to take up 4 bytes
  RepeatTile:

.PATCH 05:BB88 ; 17B98
  StoreColumnGroup:
    LSR A            ; shift once more
    STA ColumnGroup  ; and store
    LDA $6827,X      ; run hijacked code
    JMP ExitStore    ; and return to hijack point
  MaskBlock:
    LDY ColumnGroup
    CPY #$0A ; check if it's an extended group
    BCC +
    AND #$7F ; extended
    BNE ++
  + AND #$3F ; vanilla
 ++ STA BlockIndex
    JMP ExitMask
  CheckBlockRepeat:
    LDY ColumnGroup
    CPY #$0A ; check if it's an extended group
    BCS ++   ; do nothing if extended
    AND #$40 ; otherwise check repeat mask
    BEQ ++
    JMP RepeatTile
 ++ JMP NextSquare


; Overworld Screens Pointer
; -------------------------
; Expand available screens by changing the screens pointer based on screen index

.PATCH 05:A9F4 ; 16A04 Map generator checks this code
  JSR SetScreensPointer
  StoreScreensPointer:
  STX $02
  STY $03
  LDX #$00      ; existing code
  STX $06
  JMP +
  NOP           ; take up the remaining bytes
  NOP
  NOP
  NOP
  NOP
+ ASL A         ; existing code

.PATCH 05:85E3 ; 145F3 Map generator reads the pointers from this code
  SetScreensPointer:
    LDX CurrentScreen
    LDA ScreenAttrs_3,X
    AND #$7F    ; get screen code and discard mixed enemies bit
    CMP #$7C    ; compare to default max
    BCS +       ; set alternate pointer if over
    ; otherwise set the default pointer
    LDX ScreensPointer
    LDY ScreensPointer + 1
    JMP ++      ; we're done so return
    ; set the alternate pointer, defined in Screens Data section above
  + LDX ScreensPointerAlt
    LDY ScreensPointerAlt + 1
    SBC #$7C    ; readjust the index to be zero based
    ; return to store the pointer at the expected address when done
 ++ RTS


; Handle Secret Tiles
; -----------------------------


.PATCH 05:A976 ; 16986
  ; remap special blocks to their equivalent block index instead of tile index
  ; this allows the use of arbitrary block sequences for special tiles
  .db $13, $1B, $19, $14, $2C, $2C

.PATCH 05:85CA ; 145DA
  ; special block replacements (editable)
  ; use these to make secret blocks visible
  TileCodesSecretsAltOW:
    .db $13, $1B, $19, $14, $2C, $2C

.PATCH 05:AAD0 ; 16AE0
  JSR LookupSecretBlock

.PATCH 05:9388 ; 15398 Remap special blocks to their equivalent block index
  LookupSecretBlock:
    STX BlockIndex      ; store this temporarily since we'll overwrite BlockIndex at the end, map generator checks this byte
    LDY CurrentScreen
    LDA ScreenAttrs_5,Y ; load quest screen quest flag into A for comparison later
    AND #$C0            ; test if screen has general secret (bit 6 and 7 clear)
    BEQ ReplaceBlock    ; replace block if so

    LDY SaveSlot        ; otherwise load save index
    LDX QuestNum,Y      ; then get quest num
    BNE ++              ; branch if second quest

    AND #%01000000      ; otherwise test if screen has first quest secret (bit 6 set)
    BNE ReplaceBlock    ; replace block if so
   ;BEQ NormalBlock     ; otherwise use default

 ++ AND #%10000000      ; test if screen has second quest secret (bit 7 set)
    BNE ReplaceBlock    ; use normal block if not
    BEQ NormalBlock     ; use normal block if not

  ReplaceBlock:
    LDX BlockIndex
    LDY TileCodesSecretsAltOW,X
    BNE SetBlockAndReturn

  NormalBlock:
    LDX BlockIndex
    LDY TileCodesSecretsOW,X

  SetBlockAndReturn:
    STY BlockIndex
    LDA TileCodesOW - $10,Y
    RTS


; Various Hijacks and Remapping
; -----------------------------

.PATCH 05:AA8E ; 16A9E
  JSR CheckBlockIndex

.PATCH 05:ABB8 ; 16BC8
  JSR CheckBlockIndex

.PATCH 05:AC01 ; 16C11
  JSR CheckBlockIndex


; Map Block Sequences
; -------------------

.PATCH 05:AC2A ; 16C3A - Note that map generator checks these bytes
  MapBlockIdAbove0x47:
    TXA
    SBC #$48
    ASL A
    ASL A
    TAX
    LDA BlockSequencesAbove0x47,X
    STA ($00),Y
    INY
    INX
    LDA BlockSequencesAbove0x47,X
    STA ($00),Y
    TYA
    CLC
    ADC #$15
    TAY
    INX
    LDA BlockSequencesAbove0x47,X
    STA ($00),Y
    INX
    LDA BlockSequencesAbove0x47,X
    JSR WriteLastTile
    RTS
  CheckBlockIndex:
    LDX BlockIndex
    CPX #$10
    BCC +
    CPX #$48
    BCS ++
    JMP MapBlockIdAbove0x0F ; vanilla function
 ++ JMP MapBlockIdAbove0x47 ; new function
  + JMP MapBlockIdUnder0x10 ; vanilla function


;============================================================
; Tile/Block Walkability Changes
; ------------------------------
; Determine which tiles are (un)walkable
; This code applies to overworld and dungeons!
;============================================================

.PATCH 07:EDEB ; 1EDFB WalkableTileExceptions, these only apply to the overworld
  ; removes tiles 9C, AC, and AD to make them unwalkable
  .db $8D, $91, $00, $00, $00, $CC, $D2, $D5, $DF, $A0 ; (editable)

; return here after hijack
ExitCheckWalkable := $EEB7

.PATCH 07:EE72 ; 1EE82
  JMP CheckTileSwap ; was CMP CurrentTile,X
  ExitCheckTileSwap:
  BCC CheckWalkable
SwapTiles:
  STA CurrentTile,X
CheckWalkable:
  LDA CurrentTile,X
  JMP HijackCheckWalkable
  ExitHijackCheckWalkable:
  ; relocate the following check to make sure that dungeons get the hijack as well
  LDY LevelNum
  BNE ExitCheckWalkable ; exit if dungeons


; Walkable Blocks
; ---------------
; This code determines which of the new blocks are walkable

; start index for walkable tiles (editable)
; note that these are treated as hex literals, but due to
; a limitation in snarfblasm the # is added in the code
WalkableStartTileOverworld = #$50
WalkableStartTileDungeons  = #$50

; 1C7E1-1C7FF (1F unused data for DPCM sounds .dsb $1F, $FF)
; This goes beyond 1F and cuts into boss sound 1 but there is no audible difference
.PATCH 07:C7E1 ; 1C7F1
  ; at this point the tile index is loaded in A
  CheckTileSwap:
    LDY LevelNum
    BEQ ++                ; bail if overworld
    CMP WalkableStartTileDungeons ; compare A to tile index start for dungeons walkable
    BCC +                 ; jump to swap if tile is unwalkable
    LDY CurrentTile,X     ; load in current tile to compare to walkable index
    CLC
    CPY WalkableStartTileDungeons ; compare Y to tile index start for dungeons walkable
    BCS ++                ; branch if current tile out of unwalkable range
    TYA                   ; otherwise prepare for swap
  + JMP SwapTiles         ; swap 'em
 ++ CMP CurrentTile,X     ; run hijacked code
    JMP ExitCheckTileSwap ; and return to original swap logic
+++ JMP CheckWalkable     ; skip swap
  ; at this point the tile index is loaded in A
  HijackCheckWalkable:
    LDY LevelNum
    BNE ++   ; jump to underworld check if not 0
    CMP #WalkableStartTileOverworld ; tile index start for overworld walkable
    BCS +    ; it's walkable so exit
    BCC +++  ; skip underworld check
 ++ CMP #WalkableStartTileDungeons  ; tile index start for dungeons walkable
    BCS +    ; it's walkable so exit
+++ CMP #$30 ; start unwalkable
    BCC +    ; it's walkable so exit
    LDA #$FF ; set to an unwalkable index
    JMP SetTile
  + JMP ExitHijackCheckWalkable
